function varargout = ps_LST_run(job,arg) %#ok<INUSD>

%     ps_LST_run.m - Lesion segmentation by a lesion growth model, 
%     part of the LST toolbox
%     Copyright (C) 2012  Paul Schmidt
% 
%     This program is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program.  If not, see <http://www.gnu.org/licenses/>.

% add path of SPM tools

try
    addpath(fullfile(spm('dir'),'toolbox','vbm8'));
catch ME; disp(ME.message)
    error('VBM8 toolbox not found. Please install this toolbox first.');
end

% extract segmentation options
belief = job.segopts.belief;
kappa = job.segopts.initial;
phi = job.segopts.mrf;
maxiter = job.segopts.maxiter;

nimgs = numel(job.data_pve);
if nimgs ~= numel(job.data_FLAIR) 
    error('Input data must have the same length.')
end
    
% print settings
fprintf('\n')
fprintf('*--------------------------------------------------*\n')
fprintf('MS lesions segmentation by a lesion growth algorithm\n')
fprintf('Lesion belief map: ')
beliefmaps = ['GM    '; 'WM    '; 'CSF   '; 'GM+WM '; 'GM+CSF'; 'WM+CSF'; 'All   '];
fprintf('%s ', beliefmaps(belief+1,:))
fprintf('\n')
fprintf('kappa: ')
fprintf('%g ', kappa(kappa > 0))
fprintf('\n')
fprintf('phi: %g\n', phi)
fprintf('maxiter: %g\n', maxiter)
fprintf('Number of images to segment: %g\n', nimgs)
fprintf('*--------------------------------------------------*\n')
fprintf('\n')


for i = 1:nimgs
    
    % extract file path and names
    [pthPVE,namPVE,extPVE] = spm_fileparts(job.data_pve{i});  
    [pthFLAIR,namFLAIR,extFLAIR] = spm_fileparts(job.data_FLAIR{i});  
    
    % display current directory
    fprintf('Current directory: %s\n', pthPVE)
    
    %% Apply deformation field to Tissue Probability Maps
    % copy TPMs to image folder
    copyfile(fullfile(spm('dir'),'toolbox', 'LST', 'atlas*.nii'), fullfile(pthPVE, ''))
    % apply deformation fields
    namT1 = regexprep(namPVE, 'p0', '');
    % get extension of T1 image
    if exist(fullfile(pthPVE, [namT1, '.nii']),'file')
        extT1 = '.nii';
    else
        extT1 = '.img';
    end
    field1 = {fullfile(pthPVE, ['iy_', namT1, extPVE])};
    images = {fullfile(pthPVE, 'atlas_wm.nii')};
    jobDeform = struct('interp', 5, ...
                'modulate', 0);
    jobDeform.field1 = field1;
    jobDeform.images = images;
    cg_vbm8_defs(jobDeform);
    % delete TPMs
    spm_unlink(fullfile(pthPVE, 'atlas*.nii'))

     %% Coregistration of FLAIR image
     % check if coregistered bias corrected FLAIR image is allready
     % available, if not: coregister FLAIR image
     if ~exist(fullfile(pthPVE, ['r', namFLAIR, extFLAIR]),'file')
        eoptions = struct('cost_fun', 'mi', ...
                      'sep', [4 2], ...
                      'tol', [0.02 0.02 0.02 0.001 0.001 0.001], ...
                      'fwhm', [7 7]);
        roptions = struct('interp', 1, ...
                           'wrap', [0 0 0], ...
                           'mask', 0, ...
                           'prefix', 'r');
        jobCoreg = struct('ref', {{fullfile(pthPVE, [namT1, extT1])}}, ...
                       'source', {{fullfile(pthFLAIR, [namFLAIR, extFLAIR])}}, ...
                       'other', {{[]}}, ...
                       'eoptions', eoptions, ...
                       'roptions', roptions);
        spm_run_coreg_estwrite(jobCoreg);
     end
    
    %% Lesion segmentation
    fprintf('Segment MS lesions ... load images ... ')
    % Load images
    Vwm = spm_vol(fullfile(pthPVE, 'watlas_wm.nii'));
    VFLAIR = spm_vol(fullfile(pthFLAIR, ['r', namFLAIR, extFLAIR]));
    VLabel = spm_vol(fullfile(pthPVE, [namPVE, extPVE]));
    atlas_wm = spm_read_vols(Vwm);
    FLAIR = spm_read_vols(VFLAIR);    
    Label = spm_read_vols(VLabel);
    indx_total = find(Label > 0); %#ok<NASGU>
    save(fullfile(pthPVE, 'indx.mat'), 'indx_total');
    clear indx_total;
    
    % some preprocessing
    FLAIR(isnan(FLAIR)) = 0;
    spm_write_vol(VFLAIR(1), FLAIR);
    Label(isnan(Label)) = 0;
    atlas_wm(isnan(atlas_wm)) = 0;
    atlas_wm = atlas_wm .* (Label > 0);
    atlas_wm(atlas_wm < 0) = 0;
    nx = size(FLAIR, 1);
    ny = size(FLAIR, 2);
    nz = size(FLAIR, 3);
    
    fprintf('initialize lesions ...')
    FLAIR_norm = FLAIR ./ mean(FLAIR(Label > 0));
    FLAIR_norm = FLAIR_norm .* (Label > 0);
    clear FLAIR;
    
    % lesion belief map for GM
    mean_gm = mean(FLAIR_norm((Label > 1.5) & (Label < 2.5)));
    B_gm = Label .* (Label > 1.5 & Label < 2.5) .* (FLAIR_norm - mean_gm) .* atlas_wm;
    B_gm = B_gm.*(B_gm > 0);
    clear GM;

    % lesion belief map for CSF
    mean_csf = mean(FLAIR_norm(Label <= 1.5));
    B_csf = Label .* (Label < 1.5) .* (FLAIR_norm - mean_csf) .* atlas_wm;
    B_csf(B_csf < 0) = 0;

    % lesion belief map for WM
    mean_wm = mean(FLAIR_norm(Label > 2.5));
    B_wm = Label .* (Label > 2.5) .* (FLAIR_norm - mean_wm) .* atlas_wm;
    B_wm(B_wm < 0) = 0;   
    clear atlas_wm;
    % complete lesion belief map
    B = B_gm + B_wm + B_csf;
    
    switch belief
        case 0
            B_init = B_gm;
        case 1
            B_init = B_wm;
        case 2
            B_init = B_csf;
        case 3
            B_init = B_gm+B_wm;
        case 4
            B_init = B_gm+B_csf;
        case 5
            B_init = B_wm+B_csf;
        case 6
            B_init = B;
    end
    
    % smooth lesion belief maps with a simple mean filter
    indx_les = find(B_init > 0);
    neighborhood = getNeighborhood2(B_init, indx_les, 1);
    neighborhood_mean = mean(neighborhood, 1);
    B_init_mean = B_init .* 0;
    B_init_mean(indx_les) = neighborhood_mean;
    B_init_mean = B_init_mean .* B_init;
    clear B_gm; clear B_wm; clear B_csf;
    
    indx_les = find(B > 0);
    neighborhood = getNeighborhood2(B, indx_les, 1);
    neighborhood_mean = mean(neighborhood, 1);
    B_mean = B .* 0;
    B_mean(indx_les) = neighborhood_mean;
    clear [B, neighborhood_mean, neighborhood];
    
    % create independence structure
    indi_struct_brain = uint8(createIndependenceStructure(nx, ny, nz, 1) .* (FLAIR_norm > 0));
    
    for kappa_tmp = kappa(kappa > 0)
        
        % binary initial lesion map
        Lesion_init = 1 .* (B_init_mean > kappa_tmp);

        % delete initial voxels that have no neighboring Lesion
        indx = find(Label > 0);
        neighborhood = getNeighborhood2(Lesion_init, indx, 1);
        neighborhood_img = B_mean .* 0;
        neighborhood_img(indx) = sum(neighborhood > 0, 1);
        Lesion_init(neighborhood_img < 2) = 0;
        clear [neighborhood_img, neighborhood];

        if sum(Lesion_init(indx) > 0) % check if any lesions are initialized
            
            indx = uint32(find(B_mean > 0));

            Lesion_iter = Lesion_init;
            clear Lesion_init;

            iter = uint8(0);
            %disp(['... ', num2str(), ' voxels initialized ...']);
            fprintf(' grow lesions (kappa = %g):\n', kappa_tmp)
            fprintf(repmat('=', 1, maxiter))
            fprintf('\n')
            max_lesion_iter = 1;

            while iter < maxiter && max_lesion_iter > 0.01

                % display progress
                iter = iter + 1;
                fprintf('-')

                % fit a gamma distribution to the lesion class
                [a,b] = fitgamma(double(FLAIR_norm(Lesion_iter > 0.5)));
                % and a gaussian mixture model to the other classes
                dens_csfgmwm = calc_mixture(Label, FLAIR_norm, Lesion_iter);

                % get the voxels that are neighbors to lesion voxels
                neighborhood = getNeighborhood2(Lesion_iter, indx, 1);
                neighborhood_img = B_mean .* 0;
                neighborhood_img(indx) = sum(neighborhood, 1);
                neighborhood_img(Lesion_iter > 0) = 0;
                indx_tmp1 = find(neighborhood_img > 0 & indi_struct_brain == 1);
                indx_tmp2 = find(neighborhood_img > 0 & indi_struct_brain == 2);
                indx_tmp = find(neighborhood_img > 0);

                norm_les = 0 .* B_mean;
                norm_les(indx_tmp) = dgamma(FLAIR_norm(indx_tmp), a, b);
                prob = (B_mean .* norm_les) ./ dens_csfgmwm;

                % independence stucture = 1
                neighborhood = getNeighborhood2(Lesion_iter, indx_tmp1, 1);
                neighborhood = sum(neighborhood, 1);

                mf_img = 0 .* B_mean;
                mf_img(indx_tmp1) = exp(- phi .* (6 - neighborhood) + phi .* neighborhood);
                Lesion_iter = Lesion_iter + prob .* mf_img .* (neighborhood_img > 1);
                Lesion_iter(Lesion_iter > 1) = 1;

                % independence stucture = 2
                neighborhood = getNeighborhood2(Lesion_iter, indx_tmp2, 1);
                neighborhood = sum(neighborhood, 1);

                mf_img = 0 .* B_mean;
                mf_img(indx_tmp2) = exp(- phi .* (6 - neighborhood) + phi .* neighborhood);
                Lesion_iter = Lesion_iter + prob .* mf_img .* (neighborhood_img > 1);
                Lesion_iter(Lesion_iter > 1) = 1;

		max_lesion_iter = max([Lesion_iter(indx_tmp1); Lesion_iter(indx_tmp2)]);
                if isempty(max_lesion_iter)
                    max_lesion_iter = 0;
                end

            end
            fprintf('%s %d %s', ' finished after', iter, 'iterations')
            fprintf('\n')
            fprintf(repmat('=', 1, maxiter))
            fprintf('\n')

            % Which voxels are surrounded by lesions only?
            indx = find(Label > 0);
            neighborhood = getNeighborhood2(Lesion_iter, indx, 3);
            neighborhood_img = zeros(nx, ny, nz);
            neighborhood_img(indx) = sum(neighborhood > 0.01, 1);
            Lesion_iter(Lesion_iter < 1 & neighborhood_img > 18) = 1;

            % Delete all voxels that do not have a neighbor in their first order
            % neighborhood
            indx = find(Label > 0);
            neighborhood = getNeighborhood2(Lesion_iter, indx, 1);
            neighborhood_img = zeros(nx, ny, nz);
            neighborhood_img(indx) = sum(neighborhood > 0.01, 1);
            Lesion_iter(neighborhood_img == 0) = 0;

            clear neighborhood_img;
            clear norm_les;
            clear prob;
            clear mf_img;
            
        else 
            Lesion_iter = Lesion_init;
        end % END sum(Lesion_init(indx) > 0)
        Lesion_iter(isnan(Lesion_iter)) = 0;
        
        %% write segmented lesion map
        if job.output.lesions.prob
            % write probability lesion map
            fprintf('Write probability lesion map in native space ... ')
            VLesion = VLabel;
            if kappa_tmp < 1
                if kappa_tmp < .1
                    namLM = ['lesion_lbm', num2str(belief), '_00', num2str(kappa_tmp*100), '_r', namFLAIR, extFLAIR];
                else
                    namLM = ['lesion_lbm', num2str(belief), '_0', num2str(kappa_tmp*100), '_r', namFLAIR, extFLAIR];
                end
            else
                namLM = ['lesion_lbm', num2str(belief), '_', num2str(kappa_tmp*100), '_r', namFLAIR, extFLAIR];
            end
            VLesion(1).fname = fullfile(pthPVE, namLM);
            VLesion(1).dt(1) = 16;
            spm_write_vol(VLesion(1), Lesion_iter);
            
            if job.output.lesions.normalized
                % normalize probability lesion map
                fprintf('normalize probability lesion map ... ')
                images = {fullfile(pthPVE, namLM)};
                field1 = {fullfile(pthPVE, ['y_', namT1, extPVE])};
                jobDeform = struct('interp', 5, ...
                    'modulate', 0);
                jobDeform.field1 = field1;
                jobDeform.images = images;
                cg_vbm8_defs(jobDeform)
            end
            fprintf('done!\n')
        end
        
        if job.output.lesions.binary
            % write binary lesion map
            fprintf('Write binary lesion map in native space ... ')
            if job.segopts.threshold > 0
                Lesion_binary = 1 .* (Lesion_iter >= job.segopts.threshold);
            else
                Lesion_binary = 1 .* (Lesion_iter > job.segopts.threshold);
            end
            VLesion = VLabel;
            if job.segopts.threshold < 1
                if job.segopts.threshold < .1
                    txt_threshold = ['b_00', num2str(job.segopts.threshold*100)];
                else
                    txt_threshold = ['b_0', num2str(job.segopts.threshold*100)];
                end
            else
                txt_threshold = ['b_', num2str(job.segopts.threshold*100)];
            end
            if kappa_tmp < 1
                if kappa_tmp < .1
                    namLM = [txt_threshold, '_lesion_lbm', num2str(belief), '_00', num2str(kappa_tmp*100), '_r', namFLAIR, extFLAIR];
                else
                    namLM = [txt_threshold, '_lesion_lbm', num2str(belief), '_0', num2str(kappa_tmp*100), '_r', namFLAIR, extFLAIR];
                end
            else
                namLM = [txt_threshold, '_lesion_lbm', num2str(belief), '_', num2str(kappa_tmp*100), '_r', namFLAIR, extFLAIR];
            end
            VLesion(1).fname = fullfile(pthPVE, namLM);
            VLesion(1).dt(1) = 16;
            spm_write_vol(VLesion(1), Lesion_binary);
            
            if job.output.lesions.normalized
                % normalize binary lesion map
                fprintf('normalize binary lesion map ... ')
                field1 = {fullfile(pthPVE, ['y_', namT1, extPVE])};
                images = {fullfile(pthPVE, namLM)};
                jobDeform = struct('interp', 5, ...
                    'modulate', 0);
                jobDeform.field1 = field1;
                jobDeform.images = images;
                cg_vbm8_defs(jobDeform)
            end
            fprintf('done!\n')
        end
       
    end % END for kappa_tmp
    
    %% delete other images?
    if ~job.output.other
        % delete images that were needed for segmentation
        spm_unlink(fullfile(pthPVE, 'watlas*.nii'))
        spm_unlink(fullfile(pthFLAIR, ['r', namFLAIR, extFLAIR]));
        spm_unlink(fullfile(pthPVE, ['m', namT1, extPVE]));
        spm_unlink(fullfile(pthPVE, [namPVE, extPVE]));
        spm_unlink(fullfile(pthPVE, ['y_', namT1, extPVE]));
        spm_unlink(fullfile(pthPVE, ['iy_', namT1, extPVE]));
    end
    spm_unlink(fullfile(pthPVE, 'watlas*.nii'))
    clear FLAIR_norm;
    clear Label;
    
end

c = clock();
disp(['Finished successfully on ', datestr(c)]);
varargout{:} = ['Finished successfully on ', datestr(c)];

return
